/***************************************************************************
    begin                : Wed Apr 10 2002
    copyright            : (C) 2002 by Alexei Soloviev
    email                : solovjev@lynx.jinr.ru
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include<algorithm>
#include <complex>

#include"GWT.h"
#include"Signal.h"
#include"Alt.h"
#include"2dwavelet.h"  // by A.Stadnik

GWT::GWT() : 
  used_wavelet(2), even(true), 
  name(), s_name(), s_min(0), s_max(0), fixed(false) 
{}

GWT &GWT::operator<<(const int &w)
{ 
  if( w>0 && w<7 ) used_wavelet = w;
  even = w%2 == 0 ? true : false;
  return *this;
}

std::vector<double> GWT::wavelet_table(int size, double normcoef, const double &scalecoef)
{
  std::vector<double> wt(size);  // wavelet values table
  if(!size) return wt;

  switch (used_wavelet) {
  case 1:
    normcoef /= g1coefsq;
    for(int i=0; i<size; i++) wt[i] = normcoef * g1wavelet(scalecoef*i);
    break;
  case 2:
    normcoef /= g2coefsq;
    for(int i=0; i<size; i++) wt[i] = normcoef * g2wavelet(scalecoef*i);
    break;
  case 3:
    normcoef /= g3coefsq;
    for(int i=0; i<size; i++) wt[i] = normcoef * g3wavelet(scalecoef*i);
    break;
  case 4:
    normcoef /= g4coefsq;
    for(int i=0; i<size; i++) wt[i] = normcoef * g4wavelet(scalecoef*i);
    break;
  case 5:
    normcoef /= g5coefsq;
    for(int i=0; i<size; i++) wt[i] = normcoef * g5wavelet(scalecoef*i);
    break;
  case 6:
    normcoef /= g6coefsq;
    for(int i=0; i<size; i++) wt[i] = normcoef * g6wavelet(scalecoef*i);
    break;
  default:
    break;
  }

  return wt;
}

H1 GWT::transform(const H1 &h, const double &scale)
{
  int hsize = h.size();
  if(!hsize || !scale) return h;

  double normcoef = 1.0 / (hsize * sqrt (fabs (scale)));
  double scalecoef = h.xstep() / fabs (scale);
  double support = 4;  // wavelet support half-size
  int imax = min(hsize, int(support/scalecoef));

  Vector wt = wavelet_table(imax, normcoef, scalecoef);

  H1 res(hsize, h.xmin(), h.xmax(), 
	 fixed ? name.c_str() : h.title(), 
	 h.xtitle());

  for(int i=0; i<hsize; i++) {
//      for(int j=max(0,i-imax+1); j<min(hsize,i+imax); j++)
//        res[i] += h[j] * (even||i-j>=0?1:-1) * wt[abs(i-j)];
    for(int j=max(0,i-imax+1); j<i; j++)
      res[i] += h[j] * wt[i-j];
    for(int j=i; j<min(hsize,i+imax); j++)
      res[i] += h[j] * (even?1:-1) * wt[j-i];
  }

  return res;
}

H2 GWT::transform(const H1 &h, int s_bins, const double &s_min, const double &s_max)
{
  H2 res(h.size(), h.xmin(), h.xmax(), 
	 s_bins, s_min, s_max, 
	 fixed ? name.c_str() : h.title(), 
	 h.xtitle(), 
	 fixed ? s_name.c_str() : "");

  for(int i=0; i<res.ybins(); i++) res[i]=transform(h, res.i2y(i));

  return res;
}

H2  GWT::transform(const H1 &h)
{
  int sbins = h.size();
  double smin = fixed ? max(s_min,h.xstep()) : h.xstep();
  double smax = fixed ? s_max : (h.xmax()-h.xmin())/3;

  return transform(h,sbins,smin,smax);
}

H2 GWT::transform(const H2 &h, double scale)
{
  H2 res(h.xbins(), h.xmin(), h.xmax(), 
	 h.ybins(), h.ymin(), h.ymax(), 
	 h.title());

  // Altaiski's algorithm is used here:
  using Alt::fourn;
  using Alt::filterMH;

  unsigned xsize = h.xbins();
  unsigned ysize = h.ybins();

  unsigned long imagesize=xsize*ysize; // number of complex numbers
  unsigned long dsize = 2*imagesize+1; // number of real numbers
  double *ddata = new double[dsize];

  for(unsigned i=0;i<ysize;i++)
    for(unsigned j=0;j<xsize;j++)
      ddata[2*(i*xsize+j)+1] = h[i][j];

  // Step 1:
  unsigned ndim = 2;
  unsigned long *vs = new unsigned long[ndim+1];
  vs[1] = ysize;
  vs[2] = xsize;         

  fourn(ddata, vs, ndim, 1);

  complex<double> *cimage = new complex<double>[imagesize];
  for(unsigned long l=0;l<imagesize;l++)
    {
      long shift = 2*l+1;
      cimage[l] = complex<double>(ddata[shift],ddata[shift+1])/imagesize;
    }

  // Step 2:
  complex<double> *cfiltered = new complex<double>[imagesize];
  filterMH(cimage, cfiltered, xsize, ysize, scale);

  // Step 3:
  for(unsigned long l=0;l<imagesize;l++)
    {
      long shift = 2*l+1;
      ddata[shift]  = real(cfiltered[l]);
      ddata[shift+1]= imag(cfiltered[l]);
    }

  fourn(ddata, vs, ndim, -1);


  for(unsigned i=0;i<ysize;i++)
    for(unsigned j=0;j<xsize;j++)
      res[i][j] = ddata[2*(i*xsize+j)+1];


  delete [] vs;
  delete [] ddata;
  delete [] cimage;
  delete [] cfiltered;


  return res;
}

H2 GWT::transform(const H2 &h, int level, int throw_off=0)
{
  H2 res(h.xbins(), h.xmin(), h.xmax(), 
	 h.ybins(), h.ymin(), h.ymax(), 
	 h.title());

  // Stadnik's algorithm is used here:
  double **a=new double *[h.ybins()];
  for(int i=0;i<h.ybins();i++)
    a[i]=new double[h.xbins()];
  for(int i=0;i<h.ybins();i++)
    for(int j=0;j<h.xbins();j++)
      a[i][j]=h[i][j];

  // 2d decomposition "image to wavelet coefficients":
  i2w(a,h.xbins(),h.ybins(),level,8);
  // "a" now is a 2d wavelet coefficients.
  // "level" is level of decomposition, and wavelet order is 8

  // filtering:
  if(throw_off != 0) {
    double off=pow(2,throw_off);
    for(int i=0;i<h.ybins();i++)
      for(int j=0;j<h.xbins();j++)
	if(i>=h.ybins()/off || j>=h.xbins()/off)
	  a[i][j]=0;
  }
  
  // 2d reconctruction "wavelet coefficients to image":
  w2i(a,h.xbins(),h.ybins(),level,8);

  for(int i=0; i<h.ybins(); i++)
    for(int j=0; j<h.xbins(); j++)
      res[i][j]=a[i][j];

  // Just look at this:
  for(int i=0;i<h.ybins();i++) delete [] a[i];
  delete [] a;

  return res;
}

double GWT::transform(const Sample &s, const double &scale, const double &x)
{
  // Parzen-Rosenblatt estimates are used

  if(!s.size() || !scale) return 0;

  double f = 0;
  double wavecoef = s.size() * sqrt (fabs (scale));
  Sample::const_iterator p=s.begin();

  switch(used_wavelet) {
  case 1:
    while(p!=s.end()) f += g1wavelet(( x - *p++ ) / scale);
    wavecoef *= g1coefsq;
    break;
  case 2:
    while(p!=s.end()) f += g2wavelet(( x - *p++ ) / scale);
    wavecoef *= g2coefsq;
    break;
  case 3:
    while(p!=s.end()) f += g3wavelet(( x - *p++ ) / scale);
    wavecoef *= g3coefsq;
    break;
  case 4:
    while(p!=s.end()) f += g4wavelet(( x - *p++ ) / scale);
    wavecoef *= g4coefsq;
    break;
  case 5:
    while(p!=s.end()) f += g5wavelet(( x - *p++ ) / scale);
    wavecoef *= g5coefsq;
    break;
  case 6:
    while(p!=s.end()) f += g6wavelet(( x - *p++ ) / scale);
    wavecoef *= g6coefsq;
    break;
  default:
    break;
  }

  f /= wavecoef;

  return f;
}


H1 GWT::transform(const Sample &s, const double &scale, const H1 *pattern=0)
{
  if(s.empty() && pattern != 0)return *pattern;

  int bins = pattern != 0 ? pattern->xbins() : 128;
  double MIN = pattern != 0 ? pattern->xmin() : *min_element(s.begin(), s.end());
  double MAX = pattern != 0 ? pattern->xmax() : *max_element(s.begin(), s.end());

  H1 res(bins, MIN, MAX, 
	 fixed ? name.c_str() : pattern != 0 ? pattern->title() : "" , 
	 pattern != 0 ? pattern->xtitle() : "");

  if(s.empty())return res;

  for(int i=0; i<res.xbins(); i++) res[i] = transform(s, scale, res.i2x(i));

  return res;
}

H2 GWT::transform(const Sample &s, 
		  int s_bins, const double &s_min, const double &s_max, 
		  const H1 *pattern=0)
{
  H2 res(pattern != 0 ? pattern->xbins() : 128, 
	 pattern != 0 ? pattern->xmin() : *min_element(s.begin(), s.end()), 
	 pattern != 0 ? pattern->xmax() : *max_element(s.begin(), s.end()), 
	 s_bins, s_min, s_max, 
	 fixed ? name.c_str() : pattern != 0 ? pattern->title() : "" , 
	 pattern != 0 ? pattern->xtitle() : "", 
	 fixed ? s_name.c_str() : "");

  for(int i=0; i<res.ybins(); i++) res[i]=transform(s, res.i2y(i), pattern);

  return res;
}

H2 GWT::transform(const Sample &s, const H1 *pattern=0)
{
  int sbins = pattern != 0 ? pattern->size() : 128;

  double length = pattern != 0 ? pattern->xmax() - pattern->xmin() : 
    *max_element(s.begin(), s.end()) - *min_element(s.begin(), s.end());
  double division = pattern != 0 ? pattern->xstep() : length / sbins;

  double smin = fixed ? max(s_min,division) : division;
  double smax = fixed ? s_max : length / 3;

  return transform(s, sbins, smin, smax, pattern);
}

void GWT::use_parm()
{
  fixed=true;
}

void GWT::clear_parm()
{
  fixed=false;
  set_title(); set_ytitle();
  set_ymin(); set_ymax();
}
