//_______________________________________________________
// WaveDWT is the abstract base class for discrete wavelet transform  methods.
// It's include methods, like getLayer, putLyer, t2w, w2t, etc.

#include <stream.h>
#include <cmath>
#include "wavedwt.h"

#ifdef _USE_ROOT
  ClassImp(WaveDWT)
#endif
  
WaveDWT::WaveDWT(int m, int ist)
{
  // Constructor which creates object of class WSeries
  // with N=m elements of data.
  
  pws = new WSeries(m);
  pws->N = m; 
  
  IsTree=ist;
  Level=0;
  MaxLevel=0;
}

WaveDWT::~WaveDWT()
{
  // Destructor.
  delete pws;
}

void WaveDWT::t2w(const WSeries& td,int ldeep)
{
  // Virtual procedure for forward wavelet transform that 
  // take initial data from WSeries "td" and places result
  // in WSeries class.
  // Only ldeep steps of decomposition will be done.
  // By default ldeep=-1, which means do maximum number of allowed steps.
  
  Level=0;
  
  if(td.N!=pws->N) 
    {
      pws->Resize(td.N);
    }
  
  for(int q=0;q<td.N;q++) pws->data[q]=td.data[q];
  
  t2w(ldeep);
}	

void WaveDWT::t2w(int ldeep)
{
  // Virtual procedure for forward wavelet transform. 
  // Real code will appear in
  // derivative classes since it depends of the type of wavelet.
  // Only ldeep steps of decomposition will be done.
  // By default ldeep=1, which means do one step.
  
  getMaxLevel();
  
  int levs=Level;
  int levf=Level+ldeep;
  if((ldeep==-1)||(levf>MaxLevel)) levf=MaxLevel;
  
  int layf;
  for(int level=levs;level<levf;level++)
    {
      layf=IsTree?1<<level:1;
      for(int layer=0;layer<(layf);layer++)
	{
	  decompose(level,layer);
	}
      
      Level=level+1;
    }
  
  Level=levf;
}

void WaveDWT::w2t(int ldeep)
{
  // Virtual procedure for inverse wavelet transform that
  // take initial data from current WSeries class without places
  // result into object of  WSeries class.
  // Only ldeep steps of backward composition will be done.
  // By default ldeep=1, which means do one step.
  
  int levs=Level;
  int levf=Level-ldeep;
  if((ldeep==-1)||(levf<0)) levf=0;
  
  int layf;
  for(int level=levs-1;level>=levf;level--)
    {
      layf=IsTree?1<<level:1;
      for(int layer=0;layer<(layf);layer++)
	{
	  reconstruct(level,layer);
	}
      
      Level=level;
    }
  Level=levf;
}

void WaveDWT::w2t(WSeries &td, int ldeep)
{
  // Virtual procedure for inverse wavelet transform that
  // take initial data from current WSeries class and places
  // result to object td of  WSeries class.
  // Only ldeep steps of backward composition will be done.
  // By default ldeep=-1, which means do maximum allowed steps.
  
  w2t(ldeep);
  
  if(pws->N!=td.N) 
    {
      td.Resize(pws->N);
    }
  
  for(int q=0;q<pws->N;q++) td.data[q]=pws->data[q]; 
}

void WaveDWT::getLayer(WSeries &td, int k)
{
  // Get layer with given number k and put it into td array.
  // 
  
  if(k<0) { k=0; cout<<"getLayer() warning: Selected layer number 0.\n";}
  if(IsTree)
    {
      int max_layers=1<<Level;
      if(k>max_layers-1) k=max_layers-1;
      
      int m = (pws->N)>>Level;
      if (td.N!=m) td.Resize(m);
      
      int k1=getIdA(Level,k);
      for(int i=0;i<m;i++) td.data[i]=pws->data[(i<<Level)+k1];
    }
  else
    {
      char is_top=0;
      
      if(k>MaxLevel){ k=MaxLevel;cout<<"getLayer() warning: Selected layer with maximum number "<<MaxLevel<<".\n";}
      if(k==0)
	{
	  is_top=1;
	  k=Level;
	}
      
      int m = (pws->N)>>k;
      if (td.N!=m) td.Resize(m);
      
      if (is_top) 
	{
	  for(int i=0;i<m;i++) td.data[i]=pws->data[i<<k];
	}  
      else 
	{
	  k--;
	  for(int i=0;i<m;i++) td.data[i]=pws->data[((i<<1)+1)<<k];
	}
    }
}

void WaveDWT::putLayer(WSeries &td, int k)
{
  // Replace content of the layer number k with data from array td.
  //
  if(k<0) { k=0;cout<<"Selected layer number 0.\n";}
  if(IsTree)
    {
      int max_layers=1<<Level;
      if(k>max_layers-1) { k=max_layers-1; cout<<"putLayer() warning: Selected layer with maximum number "<<max_layers-1<<".\n"; }
      
      int m=(pws->N)>>Level;
      if(td.N!=m){ cout<<"putLayer() warning: number of data is not correct! Does nothing.\n";return;}
      
      int k1=getIdA(Level,k);
      for(int i=0;i<m;i++) pws->data[(i<<Level)+k1]=td.data[i];
    }
  else
    {
      char is_top=0;
      
      if(k>MaxLevel) { k=MaxLevel; cout<<"putLayer() warning: Selected layer with maximum number "<<MaxLevel<<".\n"; }
      if(k==0)
	{
	  is_top=1;
	  k=Level;
	}
      
      int m=(pws->N)>>k;
      if(td.N!=m){ cout<<"putLayer() warning: number of data is not correct! Does nothing.\n";return;}
      
      if (is_top) 
	{
	  for(int i=0;i<m;i++) pws->data[i<<k]=td.data[i];
	}  
      else 
	{
	  k--;
	  for(int i=0;i<m;i++) pws->data[((i<<1)+1)<<k]=td.data[i];
	}
    }
}

void WaveDWT::getFreqLayer(WSeries &td, int k)
{
  // The same as getLayer, but select layers in frequency order.
  
  int k1=getIdL(Level,k);
  getLayer(td,k1);	
}

void WaveDWT::putFreqLayer(WSeries &td, int k)
{
  // The same as putLayer, but put layers in frequency order.
  
  putLayer(td,getIdL(Level,k));
}

int WaveDWT::getMaxLevel()
{
  // Calculate maximal level of wavelet decomposition.
  // It depends on filter length.
  
  MaxLevel=0;
  int jc1=pws->N;
  
  for(;jc1>=nh;jc1=jc1>>1) MaxLevel++;
  return MaxLevel;
}

int WaveDWT::getIdA(int level,int layer)
{	
  // Gets wavelet layer number for given level and layer. 
  
  int IdA=0;
  for(int i=0;i<level;i++)
    {
      if((layer>>i)&1) IdA+=1<<(level-1-i);	//some operations with bits
    }
  return IdA;
}

int WaveDWT::getIdF(int level,int IdL)
{
  // Get frequency ID number for given level and layer.
  //
  
  int IdF=IdL;
  for(int i=1;i<level;i++)
    {
      int j=((1<<i)&(IdF));
      if(j)IdF=((1<<i)-1)^(IdF);
    }
  
  return IdF;
}

int WaveDWT::getIdL(int level,int IdF)
{
  // Get the layer number in binary tree for given level and frequency ID number.
  //
  
  int IdL=IdF;
  for(int i=level-1;i>=1;i--)
    {
      int j=((1<<i)&(IdL));
      if(j)IdL=((1<<i)-1)^(IdL);
    }
  
  return IdL;
}

void WaveDWT::zeroLayer(int k)
{
  // Get layer with given number k and put it to zero.
  
  if(k<0) { k=0; cout<<"zeroLayer() warning: Selected layer number 0.\n";}
  if(IsTree)
    {
      int max_layers=1<<Level;
      if(k>max_layers-1) k=max_layers-1;
      
      int m = (pws->N)>>Level;
      
      int k1=getIdA(Level,k);
      for(int i=0;i<m;i++) pws->data[(i<<Level)+k1]=0;
    }
  else
    {
      char is_top=0;
      
      if(k>MaxLevel){ k=MaxLevel;cout<<"zeroLayer() warning: Selected layer with maximum number "<<MaxLevel<<".\n";}
      if(k==0)
	{
	  is_top=1;
	  k=Level;
	}
      
      int m = (pws->N)>>k;
      
      if (is_top) 
	{
	  for(int i=0;i<m;i++) pws->data[i<<k]=0;
	}  
      else 
	{
	  k--;
	  for(int i=0;i<m;i++) pws->data[((i<<1)+1)<<k]=0;
	}
    }
}

void WaveDWT::keepLayer(int k)
{
  if(IsTree)
    {
      int max_layers=1<<Level;
      for(int i=0;i<max_layers;i++)
	if(i!=k) zeroLayer(i);
    }
  else
    {
      for(int i=0;i<=Level;i++)
	if(i!=k) zeroLayer(i);
    }
}

void WaveDWT::thresholdLayer(int k,wavereal lambda)
{
  
  if(k<0) { k=0; cout<<"thresholdLayer() warning: Selected layer number 0.\n";}
  if(IsTree)
    {
      int max_layers=1<<Level;
      if(k>max_layers-1) k=max_layers-1;
      
      int m = (pws->N)>>Level;
      
      int k1=getIdA(Level,k);
      for(int i=0;i<m;i++) 
	{
	  int ip=(i<<Level)+k1;
	  if(fabs(pws->data[ip])<lambda)pws->data[ip]=0;
	}
    }
  else
    {
      char is_top=0;
      
      if(k>MaxLevel){ k=MaxLevel;cout<<"thresholdLayer() warning: Selected layer with maximum number "<<MaxLevel<<".\n";}
      if(k==0)
	{
	  is_top=1;
	  k=Level;
	}
      
      int m = (pws->N)>>k;
      
      if (is_top) 
	{
	  for(int i=0;i<m;i++) 
	    if(fabs(pws->data[i<<k])<lambda)pws->data[i<<k]=0;
	}  
      else 
	{
	  k--;
	  for(int i=0;i<m;i++){
	    int ip=((i<<1)+1)<<k; 
	    if(fabs(pws->data[ip])<lambda)pws->data[ip]=0;
	  }
	}
    }
}

void WaveDWT::thresholdLayers(wavereal lambda)
{
  if(IsTree)
    {
      int max_layers=1<<Level;
      for(int i=0;i<max_layers;i++)
	thresholdLayer(i,lambda);
    }
  else
    {
      for(int i=0;i<=Level;i++)
	thresholdLayer(i,lambda);
    }
}

void WaveDWT::thresholdLayers(wavereal *lambdas)
{
  if(IsTree)
    {
      int max_layers=1<<Level;
      for(int i=0;i<max_layers;i++)
	thresholdLayer(i,lambdas[i]);
    }
  else
    {
      for(int i=0;i<=Level;i++)
	thresholdLayer(i,lambdas[i]);
    }
}

wavereal WaveDWT::statisticsLayer(int k,wavereal &sigma)
{
  wavereal mean=0;
  wavereal sum2=0;
  int n2;

  if(k<0) { k=0; cout<<"statisticsLayer() warning: Selected layer number 0.\n";}
  if(IsTree)
    {
      int max_layers=1<<Level;
      if(k>max_layers-1) k=max_layers-1;
      
      int m = (pws->N)>>Level; n2=m;
      
      int k1=getIdA(Level,k);
      for(int i=0;i<m;i++) 
	{
	  mean+=pws->data[(i<<Level)+k1];
	}
      mean/=m;
      for(int i=0;i<m;i++) 
	{
	  wavereal iw=pws->data[(i<<Level)+k1]-mean;
	  sum2+=iw*iw;
	}
    }
  else
    {
      char is_top=0;
      
      if(k>MaxLevel){ k=MaxLevel;cout<<"statisticsLayer() warning: Selected layer with maximum number "<<MaxLevel<<".\n";}
      if(k==0)
	{
	  is_top=1;
	  k=Level;
	}
      
      int m = (pws->N)>>k;n2=m;
      
      if (is_top) 
	{
	  for(int i=0;i<m;i++) 
	    mean+=pws->data[i<<k];
	  mean/=m;
	  for(int i=0;i<m;i++)
	    {
	      wavereal iw=pws->data[i<<k]-mean; 
	      sum2+=iw*iw;
	    }
	}  
      else 
	{
	  k--;
	  for(int i=0;i<m;i++)
	    mean+=pws->data[((i<<1)+1)<<k];
	  mean/=m;
	  for(int i=0;i<m;i++){
	    wavereal iw=pws->data[((i<<1)+1)<<k]-mean; 
	    sum2+=iw*iw;
	  }
	}
    }

  sigma=sqrt(sum2/n2);
  return mean;
}

int WaveDWT::numberofLayers()
{
  int nl=0;

  if(IsTree)
    {
      nl=1<<Level;
    }
  else
    {
      nl=Level+1;
    }
  return nl;
}
